# config.py
import os
from dataclasses import dataclass
from typing import List

from akerbp.mlops.gc.helpers import deploy_function as gc_deploy
from akerbp.mlops.gc.helpers import test_function as gc_test
from akerbp.mlops.cdf.helpers import deploy_function as cdf_deploy
from akerbp.mlops.cdf.helpers import test_function as cdf_test
from akerbp.mlops.cdf.helpers import set_up_cdf_client
from akerbp.mlops.core import helpers 
from akerbp.mlops.core import logger

logging=logger.get_logger(name='MLOps')

@dataclass
class EnvVar:
    env: str
    service_name: str
    local_deployment: str

    def __post_init__(self):
        envs = ["dev", "test", "prod"]
        if self.env not in envs:
            m = f"Environment: allowed values are {envs}"
            raise ValueError(m)

        service_names = ["training", "prediction"]
        if self.service_name not in service_names:
            m = f"Service name: allowed values are {service_names}"
            raise ValueError(m)

        local_deployments = ["True", None]
        if self.local_deployment not in local_deployments:
            m = f"Local deployment: allowed values are {local_deployments}"
            raise ValueError(m)


# Read environmental variables
getenv=EnvVar(
    env=os.getenv('ENV'), 
    service_name=os.getenv('SERVICE_NAME'), 
    local_deployment=os.getenv('LOCAL_DEPLOYMENT')
)

env = getenv.env
service_name = getenv.service_name
local_deployment = getenv.local_deployment


@dataclass
class ModelConfig:
    model_name: str
    model_file: str
    model_req_file: str
    model_artifact_folder: str
    info: str
    model_test_file: str
    deployment_platform: str

    def __post_init__(self):
        # Validation
        if self.deployment_platform not in ["cdf", "gc", "local"]:
            m = f"Deployment platform should be either 'cdf', 'gc' or 'local'"
            raise ValueError(m)

        # Derived fields
        if getenv.env == 'dev' and not getenv.local_deployment:
            self.deployment_platform = 'local'

        if self.deployment_platform == 'cdf':
            set_up_cdf_client(context='deploy')
            self.deploy_function, self.test_function = cdf_deploy, cdf_test
        elif self.deployment_platform == 'gc': 
            self.deploy_function, self.test_function = gc_deploy, gc_test
        elif self.deployment_platform == 'local': 
            self.deploy_function=self.test_function=lambda *args,**kargs: None

        self.model_code_folder = helpers.get_top_folder(self.model_file)
        self.model_import_path = helpers.as_import_path(self.model_file)
        self.test_import_path = helpers.as_import_path(self.model_test_file)

        self.files = {
            "model code": self.model_code_folder, 
            "handler": (f"akerbp.mlops.cdf","handler.py"),
            "project config": 'mlops_settings.py',
            "artifact folder": self.model_artifact_folder
        }
        if self.deployment_platform == "gc":
            files_gc = {
                "Dockerfile": ("akerbp.mlops.gc", "Dockerfile"),
                "requirements.app": ("akerbp.mlops.gc", "requirements.app"),
                "install_req_file.sh": ("akerbp.mlops.gc", "install_req_file.sh")
            }
            self.files = {**self.files, **files_gc}


@dataclass
class ProjectConfig:
    model_settings: List[ModelConfig]


def read_config():
    logging.info(f"Read project configuration")
    from mlops_settings import model_names, model_files, model_req_files
    from mlops_settings import model_artifact_folders, infos, model_test_files
    from mlops_settings import model_platforms
    settings =  zip(
        model_names, 
        model_files, 
        model_req_files, 
        model_artifact_folders,
        infos,
        model_test_files,
        model_platforms
    )
    model_settings = [ModelConfig(*s) for s in settings]
    project_settings = ProjectConfig(model_settings)
    logging.debug(f"{project_settings=}")
    
    return project_settings

project_settings = read_config()


def get_model_config(model_name):
    for setting in project_settings.model_settings:
        if model_name == setting.model_name:
            return setting
    raise ValueError(f"Didn't find settings for {model_name=}")

api_keys = dict(
   data = os.environ['COGNITE_API_KEY_DATA'],
   files = os.environ['COGNITE_API_KEY_FILES']
)


"""
deploy.py

Deploy services in either Google Cloud Run or CDF Functions. 
Model registry uses CDF Files.
""" 
import os
import traceback
import shutil

from importlib import import_module

import akerbp.mlops.model_manager as mm
from akerbp.mlops.deployment import helpers
from akerbp.mlops.core import logger, config

logging=logger.get_logger(name='MLOps')


def deploy_model(model_settings):
    env = config.env
    service_name = config.service_name

    c = model_settings
    
    logging.debug(" ")
    logging.info(f"Deploy model {c.model_name}")

    deployment_folder =f'mlops_{c.model_name}'
    function_name=f"{c.model_name}-{service_name}-{env}"

    if service_name == 'prediction':
        model_id = mm.set_up_model_artifact(c.model_artifact_folder, c.model_name)
    else:
        model_id = None

    m = "Create deployment folder and move required files/folders"
    logging.info(m)
    os.mkdir(deployment_folder)
    
    helpers.copy_to_deployment_folder(c.files, deployment_folder)

    logging.debug(f"cd {deployment_folder}")
    os.chdir(deployment_folder)

    logging.info("Write service settings file")
    mlops_settings=dict(
        model_name=c.model_name,
        model_artifact_folder=c.model_artifact_folder,
        model_import_path=c.model_import_path,
        model_code_folder=c.model_code_folder,
        model_id=model_id,
        test_import_path=c.test_import_path
    )
    # File name can't be mlops_settings.py, or there will be an
    # importing error when the service test is run (user settings <-
    # model test <- service test)
    with open('service_settings.py', 'w') as config_file:
        config_file.write(f'{mlops_settings=}')
    
    helpers.set_up_requirements(c.model_req_file)

    # * Dependencies: (user settings <- model test). Either run before
    #   going to the dep. folder or copy project config to dep. folder. 
    # * It is important to run tests after setting up the artifact
    #   folder in case it's used to test prediction service.
    # * Tests need the model requirements installed!
    logging.info(f"Run model and service tests")
    if c.model_test_file:
        helpers.run_tests(c.model_test_file)
        from akerbp.mlops.services.test_service import test_service
        test_service(c.model_name)
    else:
        logging.warning("Model test file is missing! Didn't run tests")
    # Project settings file isn't needed anymore
    logging.debug("Delete project settings from deployment folder")
    os.remove('mlops_settings.py')


    logging.info(f"Deploy {function_name} to {c.deployment_platform}")
    c.deploy_function(function_name, info=c.info[service_name])
    
    if c.test_import_path:
        logging.info("Make a test call")
        ServiceTest=import_module(c.test_import_path).ServiceTest  
        input = getattr(ServiceTest(), f"{service_name}_input")
        c.test_function(function_name, input)
    else:
        logging.warning("No test file was set up. End-to-end test skipped!")


def deploy():
    failed_models = {}
    base_path = os.getcwd()
    
    for model_settings in config.project_settings.model_settings:
        model_name = model_settings.model_name
        deployment_folder =f'mlops_{model_name}'
        try:
            deploy_model(model_settings)
        except Exception:
            trace = traceback.format_exc()
            error_message = f"Model failed to deploy!\n{trace}"
            logging.error(error_message)
            
            failed_models[model_name] = error_message
        finally:
            logging.debug(f"cd ..")
            os.chdir(base_path)
            logging.debug(f"Delete deployment folder")
            if os.path.isdir(deployment_folder):
                shutil.rmtree(deployment_folder)

    if failed_models:
        for model_name, error_message in failed_models.items():
            logging.debug(" ")
            logging.info(f"Model {model_name} failed: {error_message}")
        raise Exception("At least one model failed.")


if __name__ == '__main__':
    mm.setup()
    deploy()